local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local player = {}
local others = {}
local traffic = {}

local zones = {}
local zoneRot = {}
local altZones = {}
local altZoneRot = {}
local outZones = {}

local queueTest = 0
local speedLimit = 0
local realTimer = 0

local trafficTotal = 4 --Number of traffic cars in the prefab
local zonesTotal = 14 --Number of spawn zones in the prefab
local altZonesTotal = 14 --Number of alt spawn zones in the prefab (180 degree rotation versions of corresponding zones)
local outZonesTotal = 0 --Number of outer zones in the prefab
local lowSpeed = 16 --Minimum speed limit for randomizer, in metres per second
local highSpeed = 20 --Maximum speed limit for randomizer, in metres per second
local speedingSpeed = 28 --Excessive speed!
local forwardSpawn = 25 --Percent chance that the traffic car will spawn going forwards
local trafficAggro = 0.85 --The aggression level of the traffic cars
local playerSmash = 40000 --The amount of damage on the player car that will fail the scenario
local trafficSmash = 2000 --The amount of damage on a traffic car that will enable it to reinitialize
local nearDist = 20 --The distance of a traffic car that is considered overtaken by the player
local farDist = 320 --The distance of a traffic car that is too far away to have any fun with the player
local readyDist = 100 --The distance of a traffic car that is ready to test for teleportation, & the distance that cops can see you
local farMultiplier = 3 --Spawn test multipler for distant traffic cars; used with readyDist
local antiLagTest = 0.5 --The number of seconds before and after a teleportation to modify traffic AI, to try to prevent lag; experimental!
local randomColors = false --If true, traffic car colors can get changed to whatever!
local Traffic = {obj = nil, color = nil, pos = nil, damage = nil, state = nil, role = nil, travel = nil, zone = nil, far = nil, timer = nil}

function Traffic:new(obj, color, pos, damage, state, role, travel, zone, far, timer)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.color = color or nil --Array of colors to pick from for repainting the car; nests into RGBA arrays
	self.pos = pos or nil --Vehicle position
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, before, after
	self.role = role or nil --Purpose of the traffic car; normal, policeNormal, policeChase
	self.travel = travel or nil --Direction of travel; none, backward, forward
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.far = far or 1 --Multiplier for spawn mode; see below
	self.timer = timer or 0 --Internal logic timer
	return self
end

local function queueSpawn(aTraffic)
	local allow = true
	local randomWay = math.random(1, 100)
	local zoneIn = scenetree.findObject(tostring("spawnzone"..player.zone)) --This spawn zone is actually two away from the last spawn zone that the player passed; minimizes sudden appearing
	local zoneInPos = zoneIn:getPosition()
	local colorValues = {}
	local colorIn = aTraffic.color[math.random(1, #aTraffic.color)]
	if aTraffic.role == "policeNormal" then
		randomWay = math.random(1, forwardSpawn + 5)
	end
	for i, v in ipairs(traffic) do
		if traffic[i].zone == tostring("spawnzone"..player.zone) then
			allow = false	
		end
	end
	if allow == true then
		if randomWay <= forwardSpawn then
			aTraffic.obj:setPositionRotation(zoneInPos.x, zoneInPos.y, zoneInPos.z, 0, 0, altZoneRot[player.zone + 1].z, altZoneRot[player.zone + 1].w  + (randomWay / 1000)) --The random addition fixes a spawn rotation init bug
			aTraffic.travel = "forward"
			log("Z", "gull", aTraffic.obj.name.." teleported to "..tostring("spawnzone"..player.zone).." going away from you")
		end
		if randomWay > forwardSpawn then
			aTraffic.obj:setPositionRotation(zoneInPos.x, zoneInPos.y, zoneInPos.z, 0, 0, zoneRot[player.zone + 1].z, zoneRot[player.zone + 1].w + (randomWay / 1000)) --Incoming traffic is more exciting
			aTraffic.travel = "backward"
			log("Z", "gull", aTraffic.obj.name.." teleported to "..tostring("spawnzone"..player.zone).." going towards you")
		end
		if aTraffic.damage >= trafficSmash then
			aTraffic.obj:queueLuaCommand("vehicleResetted(false)")
		end
		if aTraffic.role == "policeChase" then
			aTraffic.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
		end
		if randomColors == true then
			colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
		else
			colorValues = {colorIn[1], colorIn[2], colorIn[3], colorIn[4]}
		end
		aTraffic.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aTraffic.obj.colorPalette0 = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aTraffic.state = "moved"
		aTraffic.timer = 0
	end
end

local function arrestPlayer()
	if player.arrestTimer % 1 == 0 then
		local countdown = 3 - math.modf(player.arrestTimer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.arrestTimer = player.arrestTimer + 0.25
	if player.arrestTimer > 3 then
		player.success = false
		scenario.finish({failed = "Busted! Get in the handcuffs."})
	end
end

local function chasePlayer(aPolice)
	aPolice.far = 2
	player.dist = (player.pos - aPolice.pos):len()
	local chaseDist = readyDist / 2
	if player.dist <= chaseDist then
		player.arrestWatch = true
		if aPolice.travel == "backward" or player.dist <= nearDist then
			if player.policeWatch == "chase" or player.policeWatch == "far" then
				aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(speedingSpeed * 1.5)..")")
				aPolice.obj:queueLuaCommand("ai.setMode('chase')")
				aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
				player.policeWatch = "near"
			end
		end
		if player.vel <= 4 then
			arrestPlayer()
		else
			player.arrestTimer = 0
		end
	else
		player.arrestWatch = false
		if player.policeWatch == "chase" or player.policeWatch == "near" then
			aPolice.obj:queueLuaCommand("ai.setMode('random')")
			aPolice.obj:queueLuaCommand("ai.driveInLane('on')")
			player.policeWatch = "far"
		end
	end
	for i, v in ipairs(traffic) do
		if traffic[i].role == "normal" then
			traffic[i].pos = traffic[i].obj:getPosition()
			local policeDist = (aPolice.pos - traffic[i].pos):len()
			if policeDist <= chaseDist and player.dist <= readyDist then
				traffic[i].obj:queueLuaCommand("ai.setSpeed(5)")
			end
			if aPolice.travel == "forward" and traffic[i].travel == "forward" and policeDist <= chaseDist then
				aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(lowSpeed)..")")
				aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
			end
		end
	end
end

local function onScenarioUIReady()
	player = {obj = nil, pos = nil, vel = nil, dist = nil, damage = nil, damageFail = nil, zone = nil, policeWatch = nil, arrestWatch = nil, arrestTimer = nil, crazy = nil, success = nil}
	player.obj = scenetree.findObject("rl_player0")
	player.damageFail = playerSmash
	player.zone = 1
	player.policeWatch = "none"
	player.arrestWatch = false
	player.arrestTimer = 0
	player.crazy = false
	player.success = true
	queueTest = -4
	realTimer = 0
	for  i = 1, trafficTotal, 1 do
		others[i] = scenetree.findObject(tostring("rl_traffic"..(i - 1)))
	end
	for  i = 1, zonesTotal, 1 do
		zones[i] = "spawnzone"..(i - 1)
	end
	for  i = 1, altZonesTotal, 1 do
		altZones[i] = "altzone"..(i - 1)
	end
	for i = 1, outZonesTotal, 1 do
		outZones[i] = "outzone"..(i - 1)
	end
	for i, v in ipairs(zones) do
		zoneRot[i] = scenetree.findObject(zones[i]):getRotation()
	end
	for i, v in ipairs(altZones) do
		altZoneRot[i] = scenetree.findObject(altZones[i]):getRotation()
	end
	if altZonesTotal < zonesTotal then
		for i = (altZonesTotal + 1), zonesTotal, 1 do
			altZoneRot[i] = zoneRot[i]
		end
	end
	for i, v in ipairs(others) do
		traffic[i] = Traffic:new(others[i], {}, nil, 0, "normal", "normal", "none", "none", farMultiplier, 0) --obj, color, pos, damage, state, role, travel, zone, far, timer
		local zonePos = scenetree.findObject(zones[i]):getPosition()
		traffic[i].obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z, 0, 0, zoneRot[i].z, zoneRot[i].w)
		traffic[i].travel = "backward"
		if i == 4 then --Hardcoded police car that appears eventually...
			zonePos = scenetree.findObject(zones[zonesTotal]):getPosition()
			traffic[i].obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z, 0, 0, altZoneRot[zonesTotal].z, altZoneRot[zonesTotal].w)
			traffic[i].state = "none"
			traffic[i].role = "policeNormal"
		end
	end
	speedLimit = math.random(lowSpeed, highSpeed) --Creates a new speed limit for all traffic cars; makes restarts more interesting but prevents traffic cars from bumping into each other
	log("Z", "gull", "Speed limit: "..speedLimit)
	traffic[1].color[1] = {0.28, 0.04, 0.02, 1.5}
	traffic[1].color[2] = {0.08, 0.08, 0.08, 1.5}
	traffic[1].color[3] = {0.32, 0.04, 0.02, 1.5}
	traffic[1].color[4] = {0.7, 0.7, 0.7, 1.2}
	
	traffic[2].color[1] = {0.02, 0.02, 0.02, 1.4}
	traffic[2].color[2] = {0.25, 0.04, 0.03, 1.4}
	traffic[2].color[3] = {0.54, 0.49, 0.32, 1.5}
	traffic[2].color[4] = {0.02, 0.03, 0.15, 1.5}
	
	traffic[3].color[1] = {0.91, 0.8, 0.07, 1.3}
	traffic[3].color[2] = {0.06, 0.65, 0.81, 1.3}
	traffic[3].color[3] = {0.19, 0.73, 0.06, 1.4}
	traffic[3].color[4] = {0.7, 0.7, 0.7, 1}
	
	traffic[4].color[1] = {0.96, 0.96, 0.96, 1.6}
end

local function onRaceStart()
	local crazyVehicles = {"ball", "barrels", "barrier", "blockwall", "bollard", "boxutility", "boxutility_large", "cannon", "caravan", "christmas_tree", "citybus", "cones", "dryvan", "flail", "flatbed", "flipramp", "haybale", "inflated_mat", "kickplate", "large_angletester", "large_bridge", "large_cannon", "large_crusher", "large_hamster_wheel", "large_metal_ramp", "large_roller", "large_spinner", "large_tilt", "metal_box", "metal_ramp", "piano", "pigeon", "roadsigns", "rocks", "rollover", "sawhorse", "semi", "streetlight", "suspensionbridge", "tanker", "tirestacks", "tirewall", "trafficbarrel", "tsfb", "tube", "van", "wall", "weightpad", "woodcrate", "woodplanks"}
	for i, v in ipairs(crazyVehicles) do
		if player.obj.jbeam == crazyVehicles[i] then
			player.crazy = true
			if player.obj.jbeam ~= "pigeon" then
				player.damageFail = 100000
			end
		end
	end
	for i, v in ipairs(traffic) do
		traffic[i].obj:queueLuaCommand("ai.setAggression("..trafficAggro..")")
		traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
		traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
		traffic[i].obj:queueLuaCommand("ai.setSpeedMode('limit')")
		traffic[i].obj:queueLuaCommand("ai.resetLearning(true)")
		traffic[i].obj:queueLuaCommand("ai.setMode('random')")
	end
end

local function onRaceTick(raceTickTime)
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	for i, v in ipairs(traffic) do
		traffic[i].pos = traffic[i].obj:getPosition()
		player.dist = (player.pos - traffic[i].pos):len()
		traffic[i].damage = map.objects[map.objectNames[traffic[i].obj.name]].damage
		if traffic[i].state == "moved" then
			if traffic[i].timer >= antiLagTest then --Any traffic timers, such as here, intentionally delay AI mode changes; experimental!
				traffic[i].obj:queueLuaCommand("ai.setMode('random')")
				traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
				traffic[i].state = "normal"
				traffic[i].far = 1
			end
			traffic[i].timer = traffic[i].timer + 0.25
		end
		if realTimer >= 30 then
			if traffic[i].role == "policeNormal" and player.policeWatch == "none" then
				traffic[i].state = "normal"
				player.policeWatch = "patrol"
			end
		end
		if player.dist <= nearDist then --If you pass a traffic car, prepare it for teleportation
			traffic[i].state = "ready"
			traffic[i].far = 1
			traffic[i].timer = 0
		end
		if player.dist <= readyDist / 2 then
			if traffic[i].role == "policeNormal" and player.policeWatch == "patrol" and (player.vel >= speedingSpeed or (player.dist <= 8 and traffic[i].damage >= 1000)) then
				traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
				traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedingSpeed * 1.5)..")")
				traffic[i].role = "policeChase"
				player.policeWatch = "chase"
			end
		end
		if player.zone > 2 then --Enables traffic spawning and other behaviors after you hit the 2nd zone
			if player.dist >= farDist and traffic[i].state == "normal" then --If you get really far away from a traffic car, prepare it for teleportation
				traffic[i].state = "ready"
				traffic[i].far = farMultiplier --Multiplier for ready distance; can prevent sudden vanishing
				traffic[i].timer = 0
			end
			if queueTest > 0 then
				if player.dist >= (readyDist * traffic[i].far) and traffic[i].state == "ready" then
					if traffic[i].timer == 0 and traffic[i].zone == "none" then
						traffic[i].obj:queueLuaCommand("ai.setMode('disabled')")
						traffic[i].far = 0
					end
					if traffic[i].timer >= antiLagTest and traffic[i].obj.name == traffic[queueTest].obj.name then
						queueSpawn(traffic[i])
						queueTest = -4 --Artificial delay for next spawn test; prevents some lag?
					end
					traffic[i].timer = traffic[i].timer + 0.25
				end
			end
		end
		if traffic[i].role == "policeChase" then
			chasePlayer(traffic[i])
		end
	end
	queueTest = queueTest + 1 --This gets updated every race tick to prevent two cars from being teleported at the same time to the same spawn zone
	realTimer = realTimer + 0.25
	if queueTest > trafficTotal then
		queueTest = 1
	end
	if realTimer == 3 then
		if player.crazy == true then
			helper.flashUiMessage("You picked this vehicle? Are you insane?", 3, false)
		end
	end
	if player.damage >= player.damageFail then
		player.success = false
		if player.arrestWatch == true then
			scenario.finish({failed = "Totaled... and busted!"})
		end
		if player.arrestWatch == false then
			scenario.finish({failed = "Totaled!"})
		end	
	end
	if realTimer > 140 then
		player.success = false
		scenario.finish({failed = "You ran out of time! Cue the sad trombone."})
	end
end

local function onRaceResult()
	if player.success == true then
		local endMessage = "Phew, you did it! Enjoy the festival!"
		if player.policeWatch == "patrol" then
			endMessage = "Phew, you did it! Plus, you're not wanted by the cops! Nice driving, champ."
		end
		if player.damage >= playerSmash / 2 then
			endMessage = "Phew, you did it! Wow, your vehicle got really beat up. Maybe you should also sign up for the demolition derby!"
		end
		if realTimer <= 100 then
			endMessage = "Phew, you did it! What a blazing fast time! Your heart must be racing."
		end
		if realTimer <= 90 then
			endMessage = "Phew, you did it! Holy fudge, what an incredibly insane time! I bet you intimidated the local street racers with that run. Fantastic driving!"
		end
		if realTimer >= 135 then
			endMessage = "Phew, you did it! Just in time, they were starting to close the doors!"
		end
		if player.crazy == true then
			endMessage = "Phew, you did it! Wait, what? With that vehicle? Wow, you're insane, but you've earned a headline in the newspaper! I'm impressed."
		end
		if player.obj.partConfig == "vehicles/semi/t65_ramplow.pc" then
			endMessage = "Phew, you did it! You're kinda evil, though. A true millionaire villain."
		end
		scenario.finish({msg = endMessage})
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then --Prevents stupid trigger confusion with bus stops
			for i, v in ipairs(zones) do
				if data.triggerName == zones[i] then
					player.zone = i + 1
				end
			end
			if player.zone > zonesTotal - 1 then --Prevents spawn zone errors
				player.zone = 1
			end
			log("Z", "gull", "Zone points: "..player.zone)
		end
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 105 then --intersections
					traffic[i].obj:queueLuaCommand("ai.setSpeed(10)")
				end
				if string.byte(tostring(data.triggerName)) == 115 then --spawn zones
					traffic[i].zone = data.triggerName
				end
				if data.triggerName == tostring("spawnzone"..(player.zone)) or data.triggerName == tostring("spawnzone"..(player.zone - 1)) then
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed(10)") --Traffic cars slow down in zones where teleports and stuff are likely
						log("Z", "gull", traffic[i].obj.name.." is aware in zone "..data.triggerName)
					end
				end
				log("Z", "gull", traffic[i].obj.name.." entered zone "..data.triggerName)
				if outZonesTotal > 0 then	
					for o, v in ipairs(badZones) do
						if data.triggerName == badZones[o] then --If a traffic car goes to where I don't want it to go, prepare it for teleportation
							traffic[i].obj:queueLuaCommand("ai.setSpeed(10)")
							traffic[i].state = "ready"
							traffic[i].far = 1
							traffic[i].timer = 0
						end
					end
				end
			end
		end
	end
	if data.event == "exit" then
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 115 then
					if traffic[i].state ~= "moved" then
						traffic[i].zone = "none"
					end
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
					end
				end
				log("Z", "gull", traffic[i].obj.name.." exited zone "..data.triggerName)
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onBeamNGTrigger = onBeamNGTrigger

return M